%%
% Matlab script that explains how to run graphsp_IP_w
%
% solves topology identification and optimal design problems 
% iterative reweighting algorithm (for topology identification)
% random disconnected graph with N = 50 nodes
%

clear, clc, close all

% number of the nodes
N = 50;
 
% regularization parameter gamma
gammaval = logspace(-3,log10(2.5),200);

% construct Lp
load pos
Length = 10;
 
pos = pos(1:N,:);
r = Length/5;
reweighted_Max_Iter = 1;
 
m = N*(N-1)/2;
 
% determine the neighbors of each node
idx = zeros(2,m);
Lp = zeros(N,N); 
stp = 0;
 
for i = 1:N,
    
    for j = i+1:N,
        % if the distance between two nodes is less than r
        % then they are neighbors.
        
        if norm( pos(i,:) - pos(j,:), 2 ) <= r,
            
            stp = stp + 1;
            idx(:,stp) = [i j]';
            Lp(i,j) = -1;
            Lp(j,i) = -1;
            
        end
        
    end
    
end
 
for p = 1:N
    
   Lp(p,p) = -sum(Lp(p,:));
   
end

idx = idx(:,1:stp);
Lp = sparse(Lp);
% done with Lp
 
% Construct incidence matrix of the controller graph
% No overlap between plan and controller
% Ecomplete = [Ep, Ec]
d = diag(Lp);
S = N^2 - N + nnz(d) - nnz(Lp);
index = zeros(S/2,2);

c = 1;

for i = 1:N,
    
    for j = i+1:N,
        
        if Lp(i,j)==0,
            
            index(c,:) = [i,j];
            
            c = c + 1;
            
        end
        
    end
    
end

% incidence matrix of the controller graph
Ec = incmat(index',N);

% number of edges in the controller graph
m = size(Ec,2);
 
% problem data
I = speye(N);

% state and control weights in LQR objective
% state weight - deviation from average
Q = eye(N) - (1/N)*ones(N,N);
% control weight
R = speye(N);

% "strengthened" state weight
Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 
 
% strengthened plant graph Laplacian
Gp = Lp + (1/N)*ones(N,N);
 
% constant term in J
const = trace(R*Lp) + 1;

% initialize feedback gain
x0 = 0.2*ones(m,1); 
 
% initialize xp
xp0 = 0.6*ones(m,1); 

% initialize xm
xm0 = 0.5*ones(m,1); 

% optimal centralized controller
outc = graphsp_IP_w(Lp,Ec,Q,R,0,ones(m,1),x0,xp0,xm0);

% performance of the optimal centralized controller
Gc = Gp + Ec*diag(outc.x)*Ec';
Jc = trace( Gc\Qp ) + diag(Ec'*R*Ec)'*outc.x - const;

% initialize the weight vector W 
% in the iterative reweighted algorithm
W = 1./(abs(outc.x) + 1.e-3); 

% maximum number of reweighted iterations
maxit_w = 3;

% initialization
x0 = outc.x; 
xp0 = outc.xp; 
xm0 = outc.xm; 

% Allocate memory for output variables
outsave.x     = zeros(maxit_w,m,length(gammaval)); % optimal vector of edge weights (before polishing)
outsave.zp    = zeros(maxit_w,m,length(gammaval)); % optimal dual vector zp
outsave.zm    = zeros(maxit_w,m,length(gammaval)); % optimal dual vector zm
outsave.xp    = zeros(maxit_w,m,length(gammaval)); % optimal primal vector xp 
outsave.xm    = zeros(maxit_w,m,length(gammaval)); % optimal primal vector xm
outsave.dg    = zeros(maxit_w,length(300),length(gammaval)); % duality gap
outsave.dresp = zeros(maxit_w,1,length(gammaval)); % dual residual plus
outsave.dresm = zeros(maxit_w,1,length(gammaval)); % dual residual minus
outsave.pres  = zeros(maxit_w,1,length(gammaval)); % dual residual
outsave.ind1  = zeros(maxit_w,1,length(gammaval)); % number of PCG affine iterations
outsave.ind2  = zeros(maxit_w,1,length(gammaval)); % number of PCG Mehrotra iterations

% save edge weights
fsave1 = zeros(m,length(gammaval));

% initial values of x, m, and Ec
fsave = x0;
m0 = m;
Ec0 = Ec;

% compute homotopy path
for k = 1:length(gammaval),
    
    gamma = gammaval(k);
      
    % iterative reweighting
    for w_step = 1:maxit_w,

        out1 = graphsp_IP_w(Lp,Ec,Q,R,gamma,W,x0,xp0,xm0);

        % W
        W = 1./(abs(out1.x) + 1.e-3);

        % initialization
        x0  = out1.x; 
        xp0 = out1.xp; 
        xm0 = out1.xm;

        % save data
        outsave.x(w_step,1:m,k)   = out1.x;
        outsave.zp(w_step,1:m,k)   = out1.zp;
        outsave.zm(w_step,1:m,k)   = out1.zm;
        outsave.xp(w_step,1:m,k)   = out1.xp;
        outsave.xm(w_step,1:m,k)   = out1.xm;
        outsave.dg(w_step,1:length(out1.dg),k)  = out1.dg;
        outsave.dresp(w_step,1:length(out1.dresp),k) = out1.dresp; 
        outsave.dresm(w_step,1:length(out1.dresm),k) = out1.dresm;
        outsave.pres(w_step,1:length(out1.pres),k) = out1.pres;
        outsave.ind1(w_step,1:length(out1.ind1),k) = out1.ind1;
        outsave.ind2(w_step,1:length(out1.ind2),k) = out1.ind2;
    
    end
       
    % greedy procedure
    f11 = out1.x;
    iii = find(abs(f11) < 1.e-6);
    f11(iii) = 0;
    indiii = setdiff(1:m,iii)';
    
    % identified incidence matrix
    Ec = Ec(:,indiii); 
    
    % initialization
    x0 = out1.x(indiii); 
    xp0 = out1.xp(indiii); 
    xm0 = out1.xm(indiii); 
    
    % weighting matrix
    W = W(indiii);
   
    k
   
    % save x in a proper way
    findex = 1;
    
    for kk = 1:m0
        
        if fsave(kk) ~= 0
            
            fsave(kk) = f11(findex);
            findex = findex+1; 
            
        end
        
    end
    
    fsave1(:,k) = fsave;

    f = f11;
    f(f==0) = [];

    if isempty(f)
        
        break;
        
    end   
    
    m = size(Ec,2);
   
end

% allocate memory
fopt = zeros(length(gammaval),m);
Jopt = zeros(length(gammaval),1);
NNZopt = zeros(length(gammaval),1);
 
% polishing step
for j = 1:length(gammaval),
    
    i = find(abs(fsave1(:,j)) < 1.e-8);
    ind = setdiff(1:m0,i)';
        
    % identified incidence matrix
    Ec_new = Ec0(:,ind);
    m2 = size(Ec_new,2);
    qvec = sum(Ec_new.*(R*Ec_new))';
        
    % initialize feedback gain
    x0 = 0.2*ones(m2,1); 

    % initialize xp
    xp0 = 0.6*ones(m2,1); 

    % initialize xm
    xm0 = 0.5*ones(m2,1); 

    % polish
    out_c = graphsp_IP_w(Lp,Ec_new,Q,R,0,ones(m2,1),x0,xp0,xm0); 
      
    f1 = out_c.x;
    
    % save the results
    fopt(j,1:length(f1)) = f1;
    
    % performance of the polished controller
    G = Gp + Ec_new*diag(f1)*Ec_new';
    Jopt(j) = trace( G\Qp ) + diag(Ec_new'*R*Ec_new)'*f1 - const;
    
    f11 = f1;
    f11(abs(f11)<10^-6) = 0;
    
    %number of nonzero elements of the centralized controller
    NNZopt(j) = nnz(f11);
 
end


% visualize results
for j = 150;
 
    i = find(abs(fsave1(:,j)) < 1.e-8);
    ind = setdiff(1:m0,i)';
    
    % identified incidence matrix
    Ec_new = Ec0(:,ind);
    
    % draw the plant (blue lines) and controller (red lines) graphs
    indexc = drawall(pos,idx,Ec_new);
    
end

figure
% number of non-zero edge weights vs gamma
semilogx(gammaval,100*(NNZopt/nnz(outc.x)), ...
         '.','LineWidth',2,'MarkerSize',25);
xlab = xlabel('regularization parameter \gamma','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('card (x)/card(xc) (%)','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','log','yscale','lin')
axis([gammaval(1) gammaval(end) 0 115*(NNZopt(1)/nnz(outc.x))])

figure
% H2 performance loss vs gamma
semilogx(gammaval,100*(Jopt/Jc - 1), ...
         'r.','LineWidth',2,'MarkerSize',25)
xlab = xlabel('regularization parameter \gamma','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('performance loss (%)','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','log','yscale','lin')
axis([gammaval(1) gammaval(end) 0 110*(Jopt(end)/Jc - 1)])

figure
% H2 performance loss vs number of non-zero edge weights
plot(100*(NNZopt/nnz(outc.x)),100*(Jopt/Jc - 1),'.k','LineWidth',2,'MarkerSize',25);
xlab = xlabel('card (x)/card(xc) (%)','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('performance loss (%)','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','lin','yscale','lin')
axis([0 115*(NNZopt(1)/nnz(outc.x)) 0 110*(Jopt(end)/Jc - 1)])

