%%
% Matlab script that explains how to 
% grow connected resistive networks using cvx
%
% Graph type: flag_p = 1; % for Erdos-Renyi network
%             flag_p = 2; % for path network
%
% Regularization parameter set to 0.5*gamma_max
%

clear, clc, close all

% number of nodes
N = 50;

% Network type
flag_p = 1; % for Erdos-Renyi network
% flag_p = 2; % for path network

if flag_p == 1
    
    % Edge probability for Erdos-Renyi network
    p = 1.05*log(N)/N;
    
    % Graph Laplacian of an Erdos-Renyi plant network
    Lp = random_graph_generator(N,p);
    eigLp = sort(eig(Lp),'ascend');

    % Check if a plant graph is connected 
    while abs(eigLp(2)) <= 1.e-12

        Lp = random_graph_generator(N,p);
        eigLp = sort(eig(Lp),'ascend');

        if abs(eigLp(2)) >= 1.e-12
            
            break;
            
        end
        
    end
 
elseif flag_p == 2,

    % Construct incidence matrix of a path graph
    Ep = sparse(toeplitz([1 -1 zeros(1,N-2)],[1 zeros(1,N-2)]));
    
    % Plant graph Laplacian
    Lp = Ep*Ep';
    
else 
    
     error('Incorrect graph type')
     
end

% Construct incidence matrix of the controller graph
% No overlap between plan and controller
% Ecomplete = [Ep, Ec]
% feel free to experiment with other choices for Ec
d = diag(Lp);
S = N^2 - N + nnz(d) - nnz(Lp);
index = zeros(S/2,2);

c = 1;

for i = 1:N,
    
    for j = i+1:N,
        
        if Lp(i,j)==0,
            
            index(c,:) = [i,j];
            
            c = c + 1;
            
        end
        
    end
    
end

% incidence matrix of the controller graph
Ec = incmat(index',N);

% number of edges in the controller graph
m = size(Ec,2);
 
% state and control weights in LQR objective
% state weight - deviation from average
Q = eye(N) - (1/N)*ones(N,N);
% control weight
R = speye(N);

% "strengthened" state weight
Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 

% Qphalf
Qphalf = sqrtm(Qp);
Qphalf = 0.5*real(Qphalf + Qphalf');

% form the vector that appears in the objective function
qvec = sum(Ec.*(R*Ec))';

% constant term in J
const = trace(R*Lp) + 1;

% compute Gp
Gp = Lp + (1/N)*ones(N,N);
invGp = (Lp + (1/N)*ones(N,N))\eye(N);
Yp = invGp*Q*invGp;

% compute gamma_max
gamma_max = max(abs(sum(Ec.*(Yp*Ec))));
% regularization parameter gamma
gamma = 0.5*full(gamma_max);

% run cvx
tic
cvx_begin sdp 
    
    variable X(N,N) symmetric
    variable xcvx(m,1)
    
    minimize (trace(X) + qvec'*xcvx + gamma*sum(xcvx))
    subject to 
    [X, Qphalf; Qphalf, Gp + Ec*diag(xcvx)*Ec'] >= 0;
    xcvx >= 0;
    
cvx_end
tcvx = toc;

% performance of sparsity-promoting controller
% obtained using cvx
Lcvx = Ec*diag(xcvx)*Ec';
Gcvx = Gp + Lcvx;
Jcvx = trace( Gcvx\(Q + Lcvx*R*Lcvx) );

% compare with IP method based on PCG algorithm pcg2
% options for IP method based on PCG algorithm pcg2
opts = struct('eps_dg',1.e-6,'eps_res',1.e-3,'maxit',100,'maxit_pcg',1000,'method','pcg2','delta_pcg_aff',.3,'delta_pcg',.3);

% code based on pcg2 
% matrix A NOT stored in memory; diagonal preconditioner
tic, 
    out3 = graphsp_IP_res(Lp,Ec,Q,R,gamma,opts);
t3 = toc;

% performance of sparsity-promoting controller
% obtained using IP method based on pcg2
Lpcg = Ec*diag(out3.x)*Ec';
G = Gp + Lpcg;
Jpcg = trace( G\(Q + Lpcg*R*Lpcg) );

% running times
% cvx vs pcg2
[tcvx t3]

% Visualize results
figure
% plot optimal vector of the edge weights (before polishing)
plot(1:m,xcvx,'*',1:m,out3.x,'ro')
xlab = xlabel('edge index','interpreter','tex');
set(xlab,'FontName','cmmr10','FontSize',26)
ylab = ylabel('edge weights','interpreter','tex');
set(ylab,'FontName','cmmr10','FontSize',26)
h = get(gcf,'CurrentAxes');
set(h,'FontName','cmr10','FontSize',20,'xscale','lin','yscale','lin')
axis([0 m+1 0 1.1*max(max([xcvx out3.x]))])
