% Sparsity-Promoting Optimal H_2 Controller for Growing Connected Resistive Networks
%
% Written by: Sepideh Hassan-Moghaddam and Mihailo R. Jovanovic
%
% May 2015
%
% Customized Algorithm based on Infeasible Primal-Dual Interior-Point Method
%
% Network Dynamics: qdot = -(Lp + Lx)*q + d
% 
% q - state of the network 
% d - white stochastic disturbance 
%
% Lp - graph Laplacian of the plant
% Lx = Ec*diag(x)*Ec' - graph Laplacian of the controller
% Ec - incidence matrix of the controller graph
% x  - vector of the edge weights
%
% Optimization Problem
%
% minimize <Qp,(Gp + Lx)^{-1}> + <R,Lx> + gamma*sum(x_i)
% 
% subject to x_i >= 0
% 
% Syntax:
%
% out = graphsp_IP_res(Lp,Ec,Q,R,gamma,opts)
%
% Inputs: (1) plant graph Laplacian Lp
%             incidence matrix of the controller graph Ec
%             state weight Q
%             controller weight R
%             sparsity-promoting parameter gamma
%
%         (2) options
%
%             opts.eps_dg    - tolerance on duality gap
%             opts.eps_res   - tolerance on dual residual
%             opts.maxit     - maximum number of IP iterations
%             opts.maxit_pcg - maximum number of PCG iterations
%             opts.method    - method used for computing search direction
%             opts.delta_pcg    - tolerance for pcg algorithm
%             opts.delta_pcg_aff   - tolerance for pcg algorithm
%
%             Search Directions Obtained Using: 
%
%             Cholesky Factorization: opts.method = 'chol'
%             PCG with storing A; diagonal preconditioner: opts.method = 'pcg1'
%             PCG without storing A; diagonal preconditioner: opts.method = 'pcg2'
%
%             If opts argument is omitted, the default values are set to
%
%             opts.eps_dg    = 1.e-6
%             opts.eps_res   = 1.e-3
%             opts.maxit     = 100
%             opts.maxit_pcg = 500
%             opts.method    = 'pcg2' 
%             opts.delta_pcg   = 0.3
%             opts.delta_pcg_aff  = 0.3
%
% Output:  out - output structure containing
%
%          out.x    - optimal vector of edge weights 
%          out.z    - optimal dual vector
%          out.s    - number of IP iterations
%          out.dg   - duality gap
%          out.res  - dual residual
%          out.ind1 - number of PCG iterations for affine search direction
%          out.ind2 - number of PCG iterations for Mehrotra search direction
%
% Additional information:
%
% http://www.umn.edu/~mihailo/software/graphsp/

function out = graphsp_IP_res(Lp,Ec,Q,R,gamma,opts) 

% Initialization
if nargin < 5
    error('At least six input arguments are required.')
elseif nargin == 5
    opts = struct('eps_dg',1.e-6,'eps_res',1.e-3,'maxit',100,'maxit_pcg',1000,'method','pcg2','delta_pcg_aff',0.3,'delta_pcg',0.3);
elseif nargin > 6
    error('Too many input arguments.')
end

% number of nodes
N = size(Ec,1);

% number of edges in the controller graph
% determines the number of optimization variables
m = size(Ec,2);

% Data preprocessing
eps_dg  = opts.eps_dg;  % tolerance on duality gap
eps_res = opts.eps_res; % tolerance on dual residual
maxit = opts.maxit;
maxit_pcg = opts.maxit_pcg;
flag = opts.method;
delta_pcg_aff = opts.delta_pcg_aff;
delta_pcg = opts.delta_pcg;

% sparse identity matrix
I = speye(N); 

% delta in IP method
delta = 3;
 
% "strengthened" plant graph Laplacian
Gp = Lp + (1/N)*ones(N,N);

% strengthened state weight 
Qp = Q + (1/N)*ones(N,N) + Lp*R*Lp; 

% % for Q = I - (1/N)*ones(N,N)
% Qp = I + Lp*R*Lp; 

% form the vector that appears in the objective function
qvec = sum(Ec.*(R*Ec))';

% initialize feedback gain
xhat = 0.5*ones(m,1); 

% compute invG
G = Gp + Ec*sparse(1:m,1:m,xhat)*Ec';
invG = G\I;

% form the dual variable from a primal one
Y = invG*Qp*invG;

% initialize the dual vector z
zhat = 1.01*abs(gamma + qvec - sum(Ec.*(Y*Ec))');

% residual vector for the dual feasibility
res_vec = gamma + qvec - sum(Ec.*(Y*Ec))' - zhat;

% Allocate memory for output variables
out.dg  = zeros(length(opts.maxit),1); % duality gap
out.res = zeros(length(opts.maxit),1); % dual residual
out.x   = zeros(m,1); % optimal vector of edge weights (before polishing)
out.z   = zeros(m,1); % optimal dual vector 
out.ind1 = zeros(length(opts.maxit),1); % number of affine PCG iterations
out.ind2 = zeros(length(opts.maxit),1); % number of Mehrotra PCG iterations

% duality gap
eta = xhat'*zhat;

% parameter mu - quantifies the current duality gap
mu = eta/m;

% division between z and x
z_over_x = zhat./xhat;

% method type
if strcmp(flag,'chol'), % Cholesky factorization

    % Cholesky factorization of the matrix A
    % A = 2*H1.*H2 + diag(z_over_x);
    Alow = chol(2*(Ec'*Y*Ec).*(Ec'*invG*Ec) + diag(z_over_x),'lower');
    
    % indices ind1 and ind2
    ind1 = 1;
    ind2 = 1;
    
elseif strcmp(flag,'pcg1'), % PCG with storing A and diagonal preconditioner
    
    % matrix A
    % A = 2*H1.*H2' + diag(z_over_x);
    A = 2*(Ec'*Y*Ec).*(Ec'*invG*Ec) + diag(z_over_x);
    % diagonal preconditioner
    diagA = diag(A);
    
elseif strcmp(flag,'pcg2'), % PCG without storing A and diagonal preconditioner
    
    % matrix A
    % A = 2*H1.*H2' + diag(z_over_x);
    % diagonal preconditioner
    % diagA = 2*diag(Ec'*Y*Ec).*diag(Ec'*invG*Ec) + z_over_x;
    diagA = 2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + z_over_x;

else 
    
    error('Wrong method')
    
end

% counting index
s = 1;

% initiate variables
out.dg(s,:)  = eta;
out.res(s,:) = norm(res_vec);
out.x = xhat;
out.z = zhat;
out.s = s;

% display
fprintf('%s %s %s %s \n\n', ' d_gap', '    p_res', '  i_1', 'i_2');  

% Customized infeasible primal-dual IP algorithm
% based on Cholesky factorization
for s = 2:maxit,
       
    % method type
    if strcmp(flag,'chol'), % Cholesky factorization

        % compute affine scaling directions
        dx_a = (Alow')\(Alow\(-res_vec - zhat));
        dz_a = -(z_over_x.*dx_a + zhat);
        
    elseif strcmp(flag,'pcg1'), % PCG with storing A and diagonal preconditioner
    
        % PCG Algorithm starts here
        % Algorithm 5.3 from Nocedal and Wright
        % same notation used
        
        % vector b
        bvec = -(res_vec + zhat);
        
        % norm of bvec
        normb = norm(bvec);
        
        % initialize dx_a
        dx_a = zeros(m,1); % use zero initial point
        
        % residual for affine scaling direction
        % rk = A*dx_a - bvec; % for dx_a \neq 0
        rk = -bvec; % for dx_a = 0
        
        % yk and pk
        yk = rk./diagA;
        pk = -yk;
        
        % A*pk
        Apk = A*pk;
        
        % affine scaling direction via PCG
        for ind1 = 1:maxit_pcg,
            
            % stopping criterion based on the relative duality gap
            if norm(rk) <= min(0.1,delta_pcg_aff*eta/normb),
                
                break;
                
            end
            
            % compute alpha_k
            alpha_k = (rk'*yk)/(pk'*Apk);
            % update dx_a
            dx_a = dx_a + alpha_k*pk;
            % compute rknew
            rknew = rk + alpha_k*Apk;
            % compute yknew
            yknew = (rknew./diagA);
            
            % update pk
            pknew = -yknew + ((rknew'*yknew)/(rk'*yk))*pk;
            
            % update rk, yk, pk, Apk
            rk = rknew;
            yk = yknew;
            pk = pknew;
            Apk = A*pk;
            
        end
        
        % update dz_a
        dz_a = -(z_over_x.*dx_a + zhat);
        
    else % PCG without storing A and diagonal preconditioner
        
        % Preconditioned conjugate gradients starts here
        % Algorithm 5.3 from Nocedal and Wright
        % same notation used
        
        % initialize dx_a
        dx_a = zeros(m,1); % xk
        
        % initialize rk - r0 = A*x0 - b
        bvec = -(res_vec + zhat);
        
        % norm of bvec
        normb = norm(bvec);
        
        % residual for affine scaling direction
        % rk = A*dx_a - bvec;
        rk = -bvec; % obtained for dx_a = 0
        
        % yk and pk
        yk = rk./diagA;
        pk = -yk;
        
        % "middle" matrix Mmid
        Mmid = Y*Ec*sparse(1:m,1:m,pk)*Ec'*invG;
        % form the vector Apk
        % Apk = 2*diag(Ec'*Mmid*Ec) + z_over_x.*pk;
        Apk = 2*sum(Ec.*(Mmid*Ec))' + z_over_x.*pk;
        
        % affine scaling direction via PCG
        for ind1 = 1:maxit_pcg,
               
            % stopping criterion based on the relative duality gap
            if norm(rk) <= min(0.1,delta_pcg_aff*eta/normb),
                
                break;
                
            end
            
            % compute alpha_k
            alpha_k = (rk'*yk)/(pk'*Apk);
            % update dx_a
            dx_a = dx_a + alpha_k*pk;
            % compute rknew
            rknew = rk + alpha_k*Apk;
            % compute yknew
            yknew = (rknew./diagA);
            
            % update pk
            pknew = -yknew + ((rknew'*yknew)/(rk'*yk))*pk;
            
            % update rk, yk, pk, Apk
            rk = rknew;
            yk = yknew;
            pk = pknew;
            
            % "middle" matrix
            Mmid = Y*Ec*sparse(1:m,1:m,pk)*Ec'*invG;
            % Apk = 2*diag(Ec'*Mmid*Ec) + z_over_x.*pk;
            Apk = 2*sum(Ec.*(Mmid*Ec))' + z_over_x.*pk;
               
       end
       
    % update dz_a
    dz_a = -(z_over_x.*dx_a + zhat);
             
    end
    
    % step-size selection for affine search direction
    if min(min(dx_a./xhat),0) < 1.e-12,
        
        alpha_x = min(1,1/abs(min(dx_a./(xhat))));
        
    else
        
        alpha_x = 1;
        
    end
    
    if min(min(dz_a./zhat),0) < 1.e-12,
        
        alpha_z = min(1,1/abs(min(dz_a./(zhat))));
        
    else
        
        alpha_z = 1;
        
    end
    
    % "centering" parameter sigma
    sigma = ((xhat + alpha_x*dx_a)'*(zhat + alpha_z*dz_a)/eta)^delta;
    
    % method type
    if strcmp(flag,'chol'), % Cholesky factorization

        % Mehrotra correction
        RHSmc = -res_vec - zhat + (sigma*mu - dx_a.*dz_a)./xhat;
        
        % search directions
        dx = (Alow')\(Alow\(RHSmc));
        % dz = -z_over_x.*dx - zhat + (sigma*mu - dx_a.*dz_a)./xhat;
        dz = -z_over_x.*dx + RHSmc + res_vec;
        
    elseif strcmp(flag,'pcg1'), % PCG with storing A; diagonal preconditioner
        
        % Mehrotra correction
        bvec_mc = (sigma*mu - dx_a.*dz_a)./xhat;
        
        % norm of bvec
        normb = norm(bvec_mc);
        
        % initialize dx_mc
        dx_mc = 0*dx_a; % dx = dx_mc + dx_a;
        
        % residual for scaling direction
        % rk = A*dx_mc - bvec_mc; % for dx_mc \neq 0
        rk = -bvec_mc; % for dx_mc = 0
        % yk and pk
        yk = rk./diagA;
        pk = -yk;
        
        % A*pk
        Apk = A*pk;
        
        % search direction via PCG
        for ind2 = 1:maxit_pcg,
                 
            % stopping criterion based on the norm of bvec
            % if norm(rk) <= eps_cg*normb,
            
            % stopping criterion based on the relative duality gap
            if norm(rk) <= min(0.1,delta_pcg*eta/normb),
                
                break;
                
            end
            
            % compute alpha_k
            alpha_k = (rk'*yk)/(pk'*Apk);
            % update dx_mc
            dx_mc = dx_mc + alpha_k*pk;
            % compute rknew
            rknew = rk + alpha_k*Apk;
            % compute yknew
            yknew = (rknew./diagA);
            
            % update pk
            pknew = -yknew + ((rknew'*yknew)/(rk'*yk))*pk;
            
            % update rk, yk, pk, Apk
            rk = rknew;
            yk = yknew;
            pk = pknew;
            Apk = A*pk;
            
        end
        
        % update dx
        dx = dx_mc + dx_a;
        % update dz
        dz = -z_over_x.*dx - zhat + bvec_mc;
        
    else % PCG without storing A; diagonal preconditioner
        
        % Mehrotra correction
        bvec_mc = (sigma*mu - dx_a.*dz_a)./xhat;
        
        % norm of bvec
        normb = norm(bvec_mc);
        
        % initialize dx_mc
        dx_mc = zeros(m,1); % dx = dx_mc + dx_a;
        
        % residual for affine scaling direction
        % rk = A*dx_mc - bvec_mc;
        rk = -bvec_mc; % for dx_mc = 0
        % yk and pk
        yk = rk./diagA;
        pk = -yk;
        
        % "middle" matrix
        Mmid = Y*Ec*sparse(1:m,1:m,pk)*Ec'*invG;
        % form the vector Apk
        % Apk = 2*diag(Ec'*Mmid*Ec) + z_over_x.*pk;
        Apk = 2*sum(Ec.*(Mmid*Ec))' + z_over_x.*pk;
        
        % search direction via PCG
        for ind2 = 1:maxit_pcg,
               
            % stopping criterion based on the relative duality gap
            if norm(rk) <= min(0.1,delta_pcg*eta/normb),
                
                break;
                
            end
            
            % compute alpha_k
            alpha_k = (rk'*yk)/(pk'*Apk);
            % update dx_mc
            dx_mc = dx_mc + alpha_k*pk;
            % compute rknew
            rknew = rk + alpha_k*Apk;
            % compute yknew
            yknew = (rknew./diagA);
            
            % update pk
            pknew = -yknew + ((rknew'*yknew)/(rk'*yk))*pk;
            
            % update rk, yk, pk, Apk
            rk = rknew;
            yk = yknew;
            pk = pknew;
            
            % "middle" matrix
            Mmid = Y*Ec*sparse(1:m,1:m,pk)*Ec'*invG;
            % form the vector Apk
            % Apk = 2*diag(Ec'*Mmid*Ec) + z_over_x.*pk;
            Apk = 2*sum(Ec.*(Mmid*Ec))' + z_over_x.*pk;
            
        end
        
        % update dx
        dx = dx_mc + dx_a;
        % update dz
        dz = -z_over_x.*dx - zhat + bvec_mc;
        
    end
        
    % step-size selection for search directions dx and dz
    if min(min(dx./xhat),0) < 1.e-12,
        
        alpha_x = min(1,1/abs(min(dx./(xhat))));
        
    else
        
        alpha_x = 1;
        
    end
    
    if min(min(dz./zhat),0) < 1.e-12,
        
        alpha_z = min(1,1/abs(min(dz./(zhat))));
        
    else
        
        alpha_z = 1;
        
    end
    
    % update xhat and zhat
    xhat = xhat + min(1,0.99*alpha_x)*dx;
    zhat = zhat + min(1,0.99*alpha_z)*dz;
     
    % duality gap
    eta = xhat'*zhat;
    
    % invG
    % G = Gp + Ec * diag(x) * Ec';
    invG = (Gp + Ec*sparse(1:m,1:m,xhat)*Ec')\I;
    
    % form the dual variable from a primal one
    Y = invG*Qp*invG;
    
    % residual vector for the dual feasibility
    res_vec = gamma + qvec - sum(Ec.*(Y*Ec))' - zhat;  
    
    % stopping criterion
    check = any( [eta > eps_dg, norm(res_vec) > eps_res] );
   
    % declare variables
    out.x = xhat;
    out.z = zhat;
    out.s = s;
    out.dg(s,:)  = eta;
    out.res(s,:) = norm(res_vec);
    out.ind1(s-1) = ind1;
    out.ind2(s-1) = ind2;
  
    % display duality gap, dual residual, ind1, and ind2
    disp([num2str(eta,'%6.1E'),'   ', ...
        num2str(norm(res_vec),'%6.1E'),'   ', ...
        num2str(ind1,'%i'),'   ', ...
        num2str(ind2,'%i')])
        
    if check == 0
        
        break;
        
    else
        
        % parameter mu - quantifies the current duality gap
        mu = eta/m;
        
        % division between z and x
        z_over_x = zhat./xhat;
        
        if strcmp(flag,'chol'),
             
            % Cholesky factorization of the matrix A
            % A = 2*H1.*H2 + diag(z_over_x);
            Alow = chol(2*(Ec'*Y*Ec).*(Ec'*invG*Ec) + diag(z_over_x),'lower');
            
        elseif strcmp(flag,'pcg1'),
             
            % matrix A
            % A = 2*H1.*H2' + diag(z_over_x);
            A = 2*(Ec'*Y*Ec).*(Ec'*invG*Ec) + diag(z_over_x);
            % diagonal preconditioner
            diagA = diag(A);
            
        else
            
            % diagonal preconditioner
            % diagA = 2*diag(Ec'*Y*Ec).*diag(Ec'*invG*Ec) + z_over_x;
            diagA = 2*sum(Ec.*(Y*Ec))'.*sum(Ec.*(invG*Ec))' + z_over_x;
            
        end

    end  
    
end

